# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_sorcery.db import meta  # noqa

from ...base import TestCase
from ...testapp.models import Business, CompositePkModel, Owner, Vehicle


class TestModelMeta(TestCase):
    def test_model_meta(self):
        info = meta.model_info(Owner)

        self.assertEqual(repr(info), "<model_info(Owner)>")
        self.assertEqual(set(info.primary_keys.keys()), {"id"})
        self.assertEqual(info.primary_keys["id"].property, Owner.id.property)

        self.assertEqual(set(info.properties.keys()), {"first_name", "last_name"})
        self.assertEqual(info.properties["first_name"].property, Owner.first_name.property)
        self.assertEqual(info.properties["last_name"].property, Owner.last_name.property)

        self.assertEqual(set(info.relationships.keys()), {"vehicles"})
        self.assertTrue(info.relationships["vehicles"].relationship is Owner.vehicles.property)

        self.assertEqual(set(info.field_names), {"id", "first_name", "last_name", "vehicles"})

        self.assertEqual(
            [attr for attr in dir(info) if not attr.startswith("__")],
            [
                "_configure",
                "_field_names",
                "composites",
                "field_names",
                "first_name",
                "get_key",
                "id",
                "identity_key_from_dict",
                "identity_key_from_instance",
                "last_name",
                "mapper",
                "model_class",
                "primary_keys",
                "primary_keys_from_dict",
                "primary_keys_from_instance",
                "properties",
                "relationships",
                "vehicles",
            ],
        )
        self.assertEqual(info.id, info.primary_keys["id"])
        self.assertEqual(info.first_name, info.properties["first_name"])
        self.assertEqual(info.vehicles, info.relationships["vehicles"])
        with self.assertRaises(AttributeError):
            info.abc

        info = meta.model_info(Business)
        self.assertEqual(info.location, info.composites["location"])
        self.assertEqual(info.other_location, info.composites["other_location"])
        self.assertNotEqual(info.location, info.other_location)

    def test_model_meta_with_mapper(self):
        mapper = Vehicle.owner.property.parent
        self.assertEqual(meta.model_info(mapper), meta.model_info(Vehicle))

    def test_primary_keys_from_dict(self):
        info = meta.model_info(Owner)

        key = info.primary_keys_from_dict({"id": 1})
        self.assertEqual(key, 1)

        info = meta.model_info(CompositePkModel)
        key = info.primary_keys_from_dict({"id": 1, "pk": 2})
        self.assertEqual(key, (1, 2))

    def test_primary_keys_from_instance(self):
        info = meta.model_info(Owner)
        instance = Owner(id=1)

        self.assertIsNone(info.primary_keys_from_instance(None))

        key = info.primary_keys_from_instance(instance)
        self.assertEqual(key, 1)

        info = meta.model_info(CompositePkModel)
        instance = CompositePkModel(id=1, pk=2)

        key = info.primary_keys_from_instance(instance)
        self.assertDictEqual(key, {"id": 1, "pk": 2})

    def test_get_key(self):
        info = meta.model_info(Owner)
        owner = Owner(id=1)

        key = info.get_key(owner)
        self.assertEqual(key, (1,))
        self.assertIsNone(info.get_key(Owner()))

    def test_identity_key_from_instance(self):
        info = meta.model_info(Owner)
        owner = Owner(id=1)

        key = info.identity_key_from_instance(owner)
        self.assertEqual(key, meta.Identity(Owner, (1,)))
        self.assertIsNone(info.identity_key_from_instance(Owner()))

    def test_identity_key_from_dict(self):
        info = meta.model_info(Owner)

        key = info.identity_key_from_dict({"id": 1})
        self.assertEqual(key, meta.Identity(Owner, (1,)))
        self.assertIsNone(info.identity_key_from_dict({}))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import warnings
from decimal import Decimal

import sqlalchemy as sa

from django import forms as djangoforms
from django.conf import settings
from django.forms import fields as djangofields
from django.utils.text import capfirst

from ... import fields as sorceryfields


class column_info(object):
    """
    A helper class that makes sqlalchemy property and column inspection easier
    """

    default_form_class = None

    __slots__ = ("property", "column", "parent")

    def __new__(cls, *args, **kwargs):
        args = list(args)
        column = kwargs.pop("column", None)
        property = kwargs.pop("property", None)

        if args:
            column = args.pop(0)
        if args:
            property = args.pop(0)

        if not isinstance(column, sa.Column):
            warnings.warn(
                "Instantiating column_info with property is deprecated, "
                "a column instance is now required and its the first argument"
            )
            column = property

        column_info_mapping = getattr(settings, "DJANGO_SORCERY", {}).get("column_info_mapping", COLUMN_INFO_MAPPING)

        override_cls = None
        for base in column.type.__class__.mro():
            if base in column_info_mapping:
                override_cls = column_info_mapping.get(base, cls)
                break

        cls = override_cls or cls
        instance = super(column_info, cls).__new__(cls)
        return instance

    def __init__(self, column, property=None, parent=None):
        if not isinstance(column, sa.Column):
            warnings.warn(
                "Instantiating column_info with property is deprecated, "
                "a column instance is now required and its the first argument"
            )
            column, property = property, column

        self.property = property
        self.column = column
        self.parent = parent

    @property
    def validators(self):
        return self.column.info.get("validators") or []

    @property
    def null(self):
        return not self.column.primary_key and self.column.nullable

    @property
    def required(self):
        return self.column.info.get("required", not self.column.nullable)

    @property
    def name(self):
        return self.property.key if self.property is not None else self.column.key

    @property
    def parent_model(self):
        return self.property.parent.class_ if self.property else None

    @property
    def help_text(self):
        return self.column.doc

    @property
    def default(self):
        return getattr(self.column.default, "arg", None)

    @property
    def choices(self):
        return getattr(self.column.type, "enum_class", None) or getattr(self.column.type, "enums", None)

    @property
    def widget(self):
        return self.column.info.get("widget_class")

    @property
    def label(self):
        label = self.column.info.get("label")
        if label:
            return label

        return capfirst(" ".join(self.name.split("_"))) if self.name else None

    @property
    def form_class(self):
        return self.column.info.get("form_class") or self.default_form_class

    def formfield(self, form_class=None, **kwargs):
        form_class = form_class or self.form_class

        if form_class is not None:
            field_kwargs = self.field_kwargs
            field_kwargs.update(kwargs)
            return form_class(**field_kwargs)

    @property
    def field_kwargs(self):
        kwargs = {"required": self.required, "validators": self.validators, "help_text": self.help_text}
        if self.default:
            if not callable(self.default):
                kwargs["initial"] = self.default

        if self.label:
            kwargs["label"] = self.label

        if self.widget:
            kwargs["widget"] = self.widget

        return kwargs

    def __repr__(self):
        return "<{!s}({!s}.{!s})>".format(
            self.__class__.__name__, self.parent_model.__name__ if self.parent_model else "<None>", self.name
        )


class string_column_info(column_info):
    default_form_class = djangofields.CharField

    @property
    def field_kwargs(self):
        kwargs = super(string_column_info, self).field_kwargs
        kwargs["max_length"] = self.column.type.length
        return kwargs


class text_column_info(string_column_info):
    @property
    def widget(self):
        return super(text_column_info, self).widget or djangoforms.Textarea


class enum_column_info(column_info):
    @property
    def form_class(self):
        form_class = super(enum_column_info, self).form_class
        if form_class:
            return form_class

        return (
            djangofields.TypedChoiceField if isinstance(self.choices, (list, set, tuple)) else sorceryfields.EnumField
        )

    @property
    def field_kwargs(self):
        kwargs = super(enum_column_info, self).field_kwargs

        if isinstance(self.choices, (list, set, tuple)):
            kwargs["choices"] = [(x, x) for x in self.choices]
        else:
            kwargs["choices"] = self.choices

        # Many of the subclass-specific formfield arguments (min_value,
        # max_value) don't apply for choice fields, so be sure to only pass
        # the values that TypedChoiceField will understand.
        for k in list(kwargs):
            if k not in (
                "choices",
                "coerce",
                "disabled",
                "empty_value",
                "enum_class",
                "error_messages",
                "help_text",
                "initial",
                "label",
                "required",
                "show_hidden_initial",
                "validators",
                "widget",
            ):
                del kwargs[k]  # pragma: nocover

        return kwargs


class numeric_column_info(column_info):
    default_form_class = djangofields.DecimalField

    @property
    def field_kwargs(self):
        kwargs = super(numeric_column_info, self).field_kwargs
        max_digits = self.column.type.precision
        decimal_places = self.column.type.scale
        if self.column.type.python_type == Decimal:
            if max_digits is not None:
                kwargs["max_digits"] = max_digits
            if decimal_places is not None:
                kwargs["decimal_places"] = decimal_places
        return kwargs


class boolean_column_info(column_info):
    @property
    def form_class(self):
        form_class = super(boolean_column_info, self).form_class
        if form_class:
            return form_class

        return djangofields.NullBooleanField if self.null else djangofields.BooleanField


class date_column_info(column_info):
    default_form_class = djangofields.DateField


class datetime_column_info(column_info):
    default_form_class = djangofields.DateTimeField


class float_column_info(column_info):
    default_form_class = djangofields.FloatField


class integer_column_info(column_info):
    default_form_class = djangofields.IntegerField


class interval_column_info(column_info):
    default_form_class = djangofields.DurationField


class time_column_info(column_info):
    default_form_class = djangofields.TimeField


COLUMN_INFO_MAPPING = {
    sa.sql.sqltypes.Enum: enum_column_info,
    sa.sql.sqltypes.String: string_column_info,
    sa.sql.sqltypes.Text: text_column_info,
    sa.sql.sqltypes.Numeric: numeric_column_info,
    sa.sql.sqltypes.Float: float_column_info,
    sa.sql.sqltypes.Integer: integer_column_info,
    sa.sql.sqltypes.Boolean: boolean_column_info,
    sa.sql.sqltypes.Date: date_column_info,
    sa.sql.sqltypes.DateTime: datetime_column_info,
    sa.sql.sqltypes.Interval: interval_column_info,
    sa.sql.sqltypes.Time: time_column_info,
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from collections import OrderedDict, namedtuple
from itertools import chain

import six

import sqlalchemy as sa

from .base import model_info_meta
from .column import column_info
from .composite import composite_info
from .relations import relation_info


Identity = namedtuple("Key", ["model", "pk"])


class model_info(six.with_metaclass(model_info_meta)):
    """
    A helper class that makes sqlalchemy model inspection easier
    """

    __slots__ = ("mapper", "properties", "_field_names", "model_class", "composites", "primary_keys", "relationships")

    def __init__(self, model):
        self.model_class = model
        self.mapper = sa.inspect(model)
        self._field_names = None
        self.properties = {}
        self.composites = {}
        self.relationships = {}
        self.primary_keys = OrderedDict()

        sa.event.listen(self.mapper, "mapper_configured", self._configure)
        self._configure(self.mapper, self.model_class)

    def _configure(self, mapper, class_):
        assert mapper is self.mapper
        assert class_ is self.model_class

        for col in self.mapper.primary_key:
            attr = self.mapper.get_property_by_column(col)
            if attr.key not in self.primary_keys:
                self.primary_keys[attr.key] = column_info(col, attr, self)

        for col in self.mapper.columns:
            attr = self.mapper.get_property_by_column(col)
            if attr.key not in self.primary_keys and attr.key not in self.properties:
                self.properties[attr.key] = column_info(col, attr, self)

        for composite in self.mapper.composites:
            if composite.key not in self.composites:
                self.composites[composite.key] = composite_info(getattr(self.model_class, composite.key))

        for relationship in self.mapper.relationships:
            if relationship.key not in self.relationships:
                self.relationships[relationship.key] = relation_info(relationship)

    def __dir__(self):
        return (
            dir(super(model_info, self))
            + list(vars(type(self)))
            + list(self.primary_keys)
            + list(self.properties)
            + list(self.composites)
            + list(self.relationships)
        )

    def __getattr__(self, name):
        if name in self.primary_keys:
            return self.primary_keys[name]
        if name in self.properties:
            return self.properties[name]
        if name in self.composites:
            return self.composites[name]
        if name in self.relationships:
            return self.relationships[name]
        return getattr(super(model_info, self), name)

    @property
    def field_names(self):
        if not self._field_names:
            self._field_names = [
                attr
                for attr in chain(
                    self.primary_keys.keys(), self.properties.keys(), self.composites.keys(), self.relationships.keys()
                )
                if not attr.startswith("_")
            ]

        return self._field_names

    def __repr__(self):
        return "<model_info(%s)>" % self.model_class.__name__

    def primary_keys_from_dict(self, kwargs):
        """
        Returns the primary key tuple from a dictionary to be used in a sqlalchemy query.get() call
        """
        pks = []

        for attr, _ in self.primary_keys.items():
            pk = kwargs.get(attr)
            pks.append(pk)

        if any(pk is None for pk in pks):
            return None

        if len(pks) < 2:
            return next(iter(pks), None)

        return tuple(pks)

    def primary_keys_from_instance(self, instance):
        """
        Return a dict containing the primary keys of the ``instance``
        """
        if instance is None:
            return None

        if len(self.primary_keys) > 1:
            return OrderedDict((name, getattr(instance, name)) for name in self.primary_keys)

        return getattr(instance, next(iter(self.primary_keys)))

    def get_key(self, instance):
        """
        Returns the primary key tuple from the ``instance``
        """
        keys = self.mapper.primary_key_from_instance(instance)
        if any(key is None for key in keys):
            return
        return keys

    def identity_key_from_instance(self, instance):
        """
        Returns the primary key tuple from the ``instance``
        """
        keys = self.get_key(instance)
        if keys is None:
            return

        return Identity(self.model_class, self.get_key(instance))

    def identity_key_from_dict(self, kwargs):
        """
        Returns identity key from a dictionary for the given model
        """
        pks = self.primary_keys_from_dict(kwargs)
        if pks is None:
            return

        return Identity(self.model_class, pks if isinstance(pks, tuple) else (pks,))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_sorcery.db import meta  # noqa

from ...base import TestCase
from ...testapp.models import Point, Vertex


class TestCompositeMeta(TestCase):
    def test_composite_meta(self):

        info = meta.model_info(Vertex)
        self.assertEqual(set(info.composites.keys()), {"start", "end"})

        start = info.composites["start"]
        end = info.composites["end"]
        self.assertEqual(set(start.properties.keys()), {"x", "y"})
        self.assertEqual(start.properties["x"].property, Vertex.x1.property)
        self.assertEqual(repr(start), "<composite_info(Point, Vertex.start)>")
        self.assertEqual(repr(end), "<composite_info(Point, Vertex.end)>")

        self.assertEqual(set(start.field_names), {"x", "y"})
        self.assertEqual(start.related_model, Point)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from collections import OrderedDict

import six

import sqlalchemy as sa

from ...utils import get_args
from .column import column_info
from .model import model_info_meta


class composite_info(six.with_metaclass(model_info_meta)):
    """
    A helper class that makes sqlalchemy composite model inspection easier
    """

    __slots__ = ("prop", "properties", "_field_names")

    def __init__(self, composite):
        self._field_names = set()
        self.prop = composite.prop

        attrs = list(sorted([k for k, v in vars(self.prop.composite_class).items() if isinstance(v, sa.Column)]))
        if not attrs:
            attrs = get_args(self.prop.composite_class.__init__)

        self.properties = OrderedDict()
        for attr, prop, col in zip(attrs, self.prop.props, self.prop.columns):
            self.properties[attr] = column_info(col, prop, self)

    @property
    def field_names(self):
        if not self._field_names:
            self._field_names.update(self.properties.keys())

            self._field_names = [attr for attr in self._field_names if not attr.startswith("_")]

        return self._field_names

    @property
    def related_model(self):
        return self.prop.composite_class

    def __repr__(self):
        return "<composite_info({!s}, {!s}.{!s})>".format(
            self.related_model.__name__, self.prop.parent.class_.__name__, self.prop.key
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from itertools import chain

from sqlalchemy.exc import NoInspectionAvailable

from django.core.exceptions import ValidationError

from ..exceptions import NestedValidationError
from ..utils import suppress
from .meta import model_info


class CleanMixin(object):
    """
    Mixin for adding django-style ``full_clean`` validation to any object.

    Base model in :py:class:`..sqlalchemy.SQLAlchemy` already uses this mixin applied.

    For example::

        class Address(db.Model):
            city = db.Column(db.String(20))
            state = db.Column(db.String(2))
            date = db.Column(db.Date())

            validators = [
                ValidateTogetherModelFields(["city", "state"]),
            ]

            def clean_date(self):
                if self.date > datetime.date.today():
                    raise ValidationError("Cant pick future date")

            def clean(self):
                if self.date.year < 1776 and self.state == "NY":
                    raise ValidationError("NY state did not exist before 1776")
    """

    def clean(self, **kwargs):
        """
        Hook for adding custom model validations before model is flushed.

        Should raise ``ValidationError`` if any errors are found.
        """

    def clean_fields(self, exclude, **kwargs):
        """
        Clean all fields on object
        """
        errors = {}

        local_remote_pairs = set()
        with suppress(NoInspectionAvailable):
            # It is possible that this could be a composite type so suppress this
            info = model_info(type(self))
            for rel in info.relationships.values():
                for col in chain(*rel.local_remote_pairs):
                    local_remote_pairs.add(col)

        for name, field in (self._get_properties_for_validation() or {}).items():
            is_blank = not bool(getattr(self, name))
            is_nullable = field.column.nullable
            is_fk = field.column in local_remote_pairs
            is_defaulted = field.column.default or field.column.server_default
            is_required = field.column.info.get("required", not is_nullable)

            # skip validation if:
            # - field is blank and either when field is nullable so blank value is valid
            # - field has either local or server default value since we assume default value will pass validation
            #   since default values are assigned during flush which as after which validation is verified
            # - field is blank and is a foreign key in a relation that will be populated by the relation
            # - field is marked as not required in column info
            is_skippable = is_blank and (is_nullable or is_defaulted or is_fk or not is_required)

            if name not in exclude and not is_skippable:

                for v in field.column.info.get("validators", []):
                    try:
                        v(getattr(self, name))
                    except ValidationError as e:
                        e = NestedValidationError(e)
                        try:
                            e.error_list
                        except AttributeError:
                            errors = e.update_error_dict(errors)
                        else:
                            errors.setdefault(name, []).extend(e.error_list)

                try:
                    getattr(self, "clean_{}".format(name), bool)()

                except ValidationError as e:
                    e = NestedValidationError(e)
                    try:
                        e.error_list
                    except AttributeError:
                        errors = e.update_error_dict(errors)
                    else:
                        errors.setdefault(name, []).extend(e.error_list)

        if errors:
            raise NestedValidationError(errors)

    def _get_properties_for_validation(self):
        """
        Needs to be implemented to return all properties for the object
        """

    def clean_nested_fields(self, exclude, **kwargs):
        """
        Clean all nested fields which includes composites
        """
        errors = {}

        for name in self._get_nested_objects_for_validation() or []:
            try:
                e = exclude.get(name, [])
            except AttributeError:
                e = []

            # only exclude when subexclude is not either list or dict
            # otherwise validate nested object and let it ignore its own subfields
            is_nestable = e and isinstance(e, (dict, list, tuple))
            if name not in exclude or is_nestable:

                try:
                    try:
                        getattr(self, name).full_clean(exclude=e, **kwargs)
                    except AttributeError:
                        pass

                except ValidationError as e:
                    e = NestedValidationError(e)
                    errors[name] = e.update_error_dict({})

        if errors:
            raise NestedValidationError(errors)

    def _get_nested_objects_for_validation(self):
        """
        Needs to be implemented to return all nested objects
        """

    def clean_relation_fields(self, exclude, **kwargs):
        """
        Clean all relation fields
        """
        visited = kwargs.pop("visited", set())
        visited.add(id(self))
        errors = {}

        for name in self._get_relation_objects_for_validation() or []:
            try:
                e = exclude.get(name, [])
            except AttributeError:
                e = []

            # only exclude when subexclude is not either list or dict
            # otherwise validate nested object and let it ignore its own subfields
            is_nestable = e and isinstance(e, (dict, list, tuple))
            if name not in exclude or is_nestable:
                value = getattr(self, name)

                if isinstance(value, (list, tuple)):
                    _errors = []

                    for i in value:
                        if id(i) not in visited:
                            try:
                                i.full_clean(exclude=e, visited=visited, **kwargs)
                            except AttributeError:
                                pass
                            except ValidationError as e:
                                e = NestedValidationError(e)
                                _errors.append(e.update_error_dict({}))

                    if _errors:
                        errors[name] = _errors

                else:
                    try:
                        if id(value) not in visited:
                            value.full_clean(exclude=e, visited=visited, **kwargs)
                    except AttributeError:
                        pass
                    except ValidationError as e:
                        e = NestedValidationError(e)
                        errors[name] = e.update_error_dict({})

        if errors:
            raise NestedValidationError(errors)

    def _get_relation_objects_for_validation(self):
        """
        Needs to be implemented to return all relation objects
        """

    def run_validators(self, **kwargs):
        """
        Check all model validators registered on ``validators`` attribute
        """
        errors = {}

        for v in getattr(self, "validators", []):
            try:
                v(self)
            except ValidationError as e:
                e = NestedValidationError(e)
                errors = e.update_error_dict(errors)

        if errors:
            raise NestedValidationError(errors)

    def full_clean(self, exclude=None, **kwargs):
        """
        Run model's full clean chain

        This will run all of these in this order:

        * will validate all columns by using ``clean_<column>`` methods
        * will validate all nested objects (e.g. composites) with ``full_clean``
        * will run through all registered validators on ``validators`` attribute
        * will run full model validation with ``self.clean()``
        * if ``recursive`` kwarg is provided, will recursively clean all relations.
          Useful when all models need to be explicitly cleaned without flushing to DB.
        """
        exclude = exclude or []
        errors = {}

        try:
            self.clean_fields(exclude=exclude, **kwargs)
        except ValidationError as e:
            e = NestedValidationError(e)
            errors = e.update_error_dict(errors)

        try:
            self.clean_nested_fields(exclude=exclude, **kwargs)
        except ValidationError as e:
            e = NestedValidationError(e)
            errors = e.update_error_dict(errors)

        try:
            self.run_validators(**kwargs)
        except ValidationError as e:
            e = NestedValidationError(e)
            errors = e.update_error_dict(errors)

        try:
            self.clean(**kwargs)
        except ValidationError as e:
            e = NestedValidationError(e)
            errors = e.update_error_dict(errors)

        if kwargs.get("recursive", False):
            try:
                self.clean_relation_fields(exclude=exclude, **kwargs)
            except ValidationError as e:
                e = NestedValidationError(e)
                errors = e.update_error_dict(errors)

        if errors:
            raise NestedValidationError(errors)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django import forms as djangoforms
from django.core import validators as django_validators

from django_sorcery.db import fields as dbfields, meta  # noqa

from ...base import TestCase
from ...testapp.models import COLORS, AllKindsOfFields, Business, Vehicle, VehicleType


class TestColumnMeta(TestCase):
    def test_deprecated_init(self):
        col = dbfields.CharField()
        info = meta.column_info(None, col)

        self.assertIs(info.column, col)
        self.assertEqual(repr(info), "<string_column_info(<None>.None)>")

    def test_column_info(self):
        info = meta.model_info(Vehicle)

        col = info.primary_keys["id"]
        self.assertEqual(col.name, "id")
        self.assertDictEqual(
            {
                "help_text": "The primary key",
                "label": "Id",
                "required": True,
                "validators": [django_validators.validate_integer],
            },
            col.field_kwargs,
        )
        self.assertIs(col.validators, col.column.info["validators"])
        self.assertTrue(col.required)

    def test_default_initial(self):
        info = meta.model_info(Business)

        col = info.employees
        self.assertDictEqual(
            {
                "help_text": None,
                "initial": 5,
                "label": "Employees",
                "required": True,
                "validators": [django_validators.validate_integer],
            },
            col.field_kwargs,
        )

    def test_column_info_enum(self):
        info = meta.model_info(Vehicle)
        col = info.type
        self.assertDictEqual(
            {"choices": VehicleType, "help_text": None, "label": "Type", "required": True, "validators": []},
            col.field_kwargs,
        )
        self.assertTrue(col.required)

    def test_column_info_validators(self):
        info = meta.model_info(Vehicle)
        col = info.properties["msrp"]
        validator = col.column.info["validators"][0]
        self.assertIsInstance(validator, django_validators.DecimalValidator)
        self.assertDictEqual(
            {
                "decimal_places": 2,
                "help_text": None,
                "label": "Msrp",
                "max_digits": 10,
                "required": False,
                "validators": [validator],
            },
            col.field_kwargs,
        )

    def test_column_info_label_and_text(self):
        column = meta.column_info(dbfields.TextField(label="dummy"))
        self.assertDictEqual(
            column.field_kwargs,
            {
                "help_text": None,
                "label": "dummy",
                "max_length": None,
                "required": False,
                "validators": [],
                "widget": djangoforms.Textarea,
            },
        )

    def test_choice_enum(self):
        info = meta.model_info(Vehicle)
        col = info.paint
        self.assertDictEqual(
            {
                "choices": [(x, x) for x in COLORS],
                "help_text": None,
                "label": "Paint",
                "required": False,
                "validators": [],
            },
            col.field_kwargs,
        )
        self.assertEqual(col.parent_model, Vehicle)
        self.assertEqual(repr(col), "<enum_column_info(Vehicle.paint)>")

    def test_plain_sqla(self):
        info = meta.model_info(AllKindsOfFields)
        col = info.decimal
        self.assertDictEqual(
            {"help_text": None, "label": "Decimal", "required": False, "validators": []}, col.field_kwargs
        )
